#! /usr/ug/bin/perl -w
#
# kamaileon -- an anonymous email delivery tool
#              by Bret Victor,  November 2001
#              bret@ugcs.caltech.edu
#              http://www.ugcs.caltech.edu/~bret
#

use IO::Socket;

$version = "1.04";
$defaulteditor = $ENV{EDITOR} || "emacs";
$defaultmailserver = "mail";
$defaultfilename = "kamaileon_message";
$noraw = 0;
$clear = `clear` || "";


sub help { print <<"_eot_";
kamaileon v$version -- Bret Victor, 1/16/02                  (c) 2002 Bret Victor

usage:  kamaileon [options]

         -help          you are here.
         -tutorial      use the tutorial template instead of the brief one.
         -e EDITOR      use the specified editor (default is $defaulteditor).
         -k             use backup file.  Edit an existing "$defaultfilename"
                          file instead of an empty template.
         -i FILENAME    import a message that was saved or exported from
                          another email client
         -n NUMBER      with the -i option, specifies which message to import
                          from a mailbox or mail "folder".  If NUMBER is
                          negative, starts counting from the end.  Default
                          is -1 (ie, the last message in the file).
         -f FILENAME    non-interactive mode.  Read and process the file
                          without starting up an editor.
         -p             piped non-interactive mode.  Read and process stdin.
         -m HOSTNAME    send via this server (default is $defaultmailserver).
         -v             verbose mode, display server transaction.
         -d             debug mode, don't actually send the message.

_eot_
}


# parse command line options

%options = ();

while (@ARGV and $ARGV[0] =~ /^-+(.)/) {
    my $option = $1;
    my $opstring = shift;

    if ($option =~ /^h/i) { help();  exit 1 }
    elsif ($option =~ /^t/i) { $options{tutorial} = 1 }
    elsif ($option eq 'e') { $options{editor} = shift }
    elsif ($option eq 'k') { $options{usebackupfile} = 1 }
    elsif ($option eq 'i') { $options{importfile} = shift }
    elsif ($option eq 'n') { $options{importnum} = shift }
    elsif ($option eq 'f') { $options{filename} = shift;
                             $options{noninteractive} = 1 }
    elsif ($option eq 'p') { $options{filename} = "-";
                             $options{noninteractive} = 1 }
    elsif ($option eq 'm') { $options{mailserver} = shift }
    elsif ($option eq 'd') { $options{debug} = 1;
	                     $options{verbose} = 1 }
    elsif ($option eq 'v') { $options{verbose} = 1 }
    else {
	help();
	print "unrecognized option:  $opstring\n";
        exit 1;
    }
}

$options{mailserver} ||= $defaultmailserver;
$options{filename} ||= $defaultfilename;
$options{editor} ||= $defaulteditor;
$options{importnum} ||= -1;

unless ($options{noninteractive}) {
    $editorpath = `which $options{editor}`;
    chomp $editorpath;
    unless ($editorpath) { 
	print "Unable to find the editor program \"$options{editor}\".\n";
	exit 1;
    }
    
    $alteditor = ($options{editor} =~ /emacs/i) ? "pico" : "emacs";
    
    unless ($options{usebackupfile} and -f $options{filename}) {
	unless (open OUT, ">$options{filename}") {
	    print "Unable to open \"$options{filename}\" for writing.\n";
	    exit 1;
	}
	unless ($options{importfile}) {
	    print OUT template();
	} else {
	    my @messagepositions = ();
	    my %import = ();
	    unless (open IMPORT, $options{importfile}) {
		print "Unable to open \"$options{importfile}\" for reading.\n";
		exit 1;
	    }

	    push @messagepositions, 0 if defined
		( $_ = <IMPORT> ) and not /^From MAILER-DAEMON /;
	    while (<IMPORT>) { 
		push @messagepositions, tell IMPORT if /^From /;
	    }

	    if (abs($options{importnum}) <= @messagepositions) {
		if ($options{importnum} < 0) { 
		    $options{importnum} += 1 + @messagepositions;
		}
		seek IMPORT, $messagepositions[$options{importnum} - 1], 0;

		my $currentcomponent = "";
		while (<IMPORT>) {
		    chomp;
		    if ($currentcomponent eq 'body') {
			last if /^From /;
			$import{$currentcomponent} .= "\n$_";
		    } elsif (/^([-\w]+):\s*(.*?)\s*$/) {
			$currentcomponent = lc $1;
			$import{$currentcomponent} = $2;
		    } elsif ($currentcomponent) {
			if (/\S/) { $import{$currentcomponent} .= "\n$_"; }
			else { $currentcomponent = "body" }
		    }
		}
	    }
	    close IMPORT;
	    print OUT template(\%import);
	}
	close OUT;
    }
}

EDITANDPARSE:  while (1) { 
    eval { 
	my %message = ();

	unless ($options{noninteractive}) {
	    system "$editorpath", $options{filename};
	    print "\n", $clear;
	}
	
	open IN, $options{filename} or die 
	    "!Unable to open \"$options{filename}\" for reading.\n";

	my $currentcomponent = "";
	while (<IN>) {
	    next if /^\#/;
	    chomp;
	    if ($currentcomponent eq 'body') {
		$message{$currentcomponent} .= "\n$_";
	    } elsif (/^([-\w]+):\s*(.*?)\s*$/) {
		$currentcomponent = lc $1;
		$message{$currentcomponent} = $2;
	    } elsif ($currentcomponent and /\S/) {
                unless ($message{$currentcomponent} =~ /\,\s*$/) {
                    $message{$currentcomponent} .= ",";
                }  
		$message{$currentcomponent} .= "\n   $_";
	    }
	}
	
	unless (exists $message{sendto}) {
	    die "You did not specify any recipients in the SendTo line.\n";
	}
	@recipients = ();
	%recipientfiles = ();
	grabrecipients($message{sendto});
	unless (@recipients) {
	    unless (exists $message{body} and $message{body} =~ /\S/) {
		die "!No message was entered.\n";
	    }
	    die "You did not specify any recipients in the SendTo line.\n";
	}

	unless ($message{from}) {
	    die "You did not specify a sender in the From line.\n";
	}
	if ($message{from} =~ /</ and $message{from} !~	/<[^<,\s]+>/) {
	    die "You did not close your pointy brackets properly in the ".
		"From line:\nFrom: $message{from}\n";
	}

	$sendername = ($message{from} =~ /([^<>\s]+)\@/) ? $1 : "nobody";
	$senderdomain = ($message{from} =~ /\@([^<>\s]+)/ and
			 defined gethostbyname lc $1) ? lc $1 : "nowhere.com";
	
	$messagedata = "From: $message{from}\n";
	if ($message{to}) {
	    my $to = $message{to};
	    if (lc($to) eq 'honest') { $to = join(",\n   ", @recipients) }
	    $messagedata .= "To: $to\n";
	}
	if ($message{cc}) {
	    my $cc = $message{cc};
	    $messagedata .= "Cc: $cc\n";
	}
	if ($message{subject}) {
	    my $subject = $message{subject};
	    $subject =~ s/\n/ /g;
	    $messagedata .= "Subject: $subject\n";
	}
	$messagedata .= "MIME-Version: 1.0\n";
	unless ($message{"content-type"}) {
	    $message{"content-type"} = "TEXT/PLAIN; charset=US-ASCII";
	}
	$messagedata .= "Content-Type: " . $message{"content-type"} . "\n\n";

	if ($message{body}) {
	    my $body = $message{body};
	    $body =~ s/^\n+//;
	    $body =~ s/^(\.\s*)$/ $1/mg;
	    $messagedata .= "$body\n";
	}

	unless ($options{noninteractive}) {
            open CONFIRM, "|more" or open CONFIRM, ">-";
	    print CONFIRM "SendTo: ";
	    if (@recipients <= 5) {
		foreach (0..4) {
		    print CONFIRM "$recipients[$_]\n", ' 'x8;
		    last if $_ == $#recipients;
		}
	    } else { print CONFIRM "(", scalar @recipients, " names)\n" }
	    print CONFIRM "\n", $messagedata;
	    close CONFIRM;

	    print "\nDo you want to (s)end this message, (e)dit the message, ".
		"or (q)uit?  ";
	    my $char;
	    do { $char = getsinglechar() } while ($char !~ /[SEQ]/);
	    print $char, "\n\n";
	    if ($char eq 'E') { die ":\n" }
	    if ($char eq 'Q') { die "*\n" }
	}	

	unless ($options{debug}) {
	    print "Connecting to server...\n" unless $options{noninteractive};
	    $remote = IO::Socket::INET->new( Proto    => "tcp",
					     PeerAddr => $options{mailserver},
					     PeerPort => 25,
					     );
	    unless ($remote) {
		die "!Unable to connect to ".
		    "mailserver \"$options{mailserver}\".\n";
	    }
	    $remote->autoflush(1);
	    print "Sending message...\n" unless $options{noninteractive};
	}

	checkresponse();
	sendline("HELO $senderdomain\n");
	checkresponse();
	sendline("MAIL From: <$sendername\@$senderdomain>\n");
	checkresponse();
	foreach $rcpt (@recipients) {
	    sendline("RCPT To: <$rcpt>\n");
	}
	foreach $rcpt (@recipients) {
	    checkresponse();
	}
	sendline("DATA\n");
	checkresponse(3);
	sendline($messagedata);
	sendline(".\n");
	checkresponse();
	sendline("QUIT\n");
	checkresponse();

	close $remote unless $options{debug};
	    
	unless ($options{noninteractive}) {
	    print "Message sent.\n" unless $options{debug};
	    unlink $options{filename};
	}
    };

    unlink "$options{filename}~" unless $options{noninteractive};
    last unless ($@);
    if ($@ =~ /^\*/) { exit 0 }
    if ($@ =~ /^!/) { print substr($@, 1);  exit 1 }
    if ($options{noninteractive}) { print $@;  exit 1 }
    if ($@ =~ /^:/) { next EDITANDPARSE }
    print $@, "Would you like to (e)dit the message or (q)uit?  ";
    my $char;
    do { $char = getsinglechar() } while ($char !~ /[EQ]/);
    print $char, "\n\n";
    exit 0 if $char eq 'Q';
}

exit 0;


sub checkresponse {
    return 1 if $options{debug};
    my $number = $_[0] || "2";
    my $response = <$remote>;
    print $response if $options{verbose};
    return 1 if (substr($response, 0, 1) eq $number);
    die "The mailserver returned an error:\n$response\n";
}

sub sendline {
    my ($line) = @_;
    print $line if $options{verbose};
    print $remote $line unless $options{debug};
}

sub grabrecipients {
    my ($tolist, $filename) = @_;
    my @list = split /[\n\r,]+/, $tolist;
    foreach $entry (@list) {
	$entry =~ s/\s//g;
	if ($entry =~ /<(.*)>/) { $entry = $1 }
	if ($entry =~ /.\@./) {
	    push @recipients, $entry;
	} elsif (not exists $recipientfiles{$entry}) {
	    $recipientfiles{$entry} = 1;
	    my $filestring = readfile($entry);
	    unless (defined $filestring) {
		my $location = (defined $filename) ? 
		    "the file \"$filename\"" : "the SendTo line";
		die "In $location, I found the entry \"$entry\".\n" .
		    "It's not an email address and I can't find a file by " .
			"that name.\n";
	    }
	    grabrecipients($filestring, $entry);
	}
    }
}


sub getsinglechar {
    if ($noraw) { return uc(substr(<STDIN>,0,1)) }
    system 'stty', 'raw', '-echo';
    my $char = getc(STDIN);
    system 'stty', 'cooked', 'echo';
    return uc $char;
}


sub readfile {
    my ($filename) = @_;
    my $filestring = "";

    open READFILE, $filename or return undef;
    while (<READFILE>) { $filestring .= $_ }
    close READFILE;
    return $filestring;
}


sub template {
    my ($import) = @_;
    $_ = $options{editor};
    my $quitstring = /^emacs/   && 'control-x control-c' ||
	             /^pico/    && 'control-x' ||
                     /^n?vi/    && '":q"' ||
                     /^n?e[xd]/ && '"q"' ||
                     /^joe/     && 'control-c' ||
			           '';

    my $template = <<"_eot_";
# This is kamaileon v$version, using the $options{editor} editor.
_eot_
    $template .= <<"_eot_" if $quitstring;
# To quit this editor, type $quitstring.
_eot_

    if ($import) {
	my $notice = "";
	if (keys %$import) {
	    $notice = "# This message was imported from message " .
		"$options{importnum} in the file \"$options{importfile}\".";
	} else {
	    $notice = "# Message $options{importnum} could not be found in " .
		"the file \"$options{importfile}\".";
	}
	if (length($notice) > 73) { $notice =~ s/file/\n\# file/ }

	$template .= "$notice\n\n";
	$template .= "SendTo: " . ($$import{to} || "");
	$template .= "\n        " if ($$import{to} && $$import{cc});
	$template .= ($$import{cc} || "") . "\n";
        $template .= "From: " . ($$import{from} || "") . "\n";
        $template .= "To: " . ($$import{to} || "") . "\n";
        $template .= "Cc: " . $$import{cc} . "\n" if $$import{cc};
        $template .= "Subject: " . ($$import{subject} || "") . "\n";
        $template .= "Content-Type: " . $$import{"content-type"} . "\n"
	 	if $$import{"content-type"};
        $template .= "\nBody:\n" . ($$import{body} || "") . "\n";

    } elsif ($options{tutorial}) {
	$template .= <<"_eot_";
# For a list of command line options, run:
#   kamaileon -help
# In particular, to use a different editor, use the -e option.  For example:
#   kamaileon -e $alteditor
#
# When you have finished filling in the lines below, save this file and
# quit the editor, and you will be given the option of sending the message.

# Who do you want the email to be sent to?  You can give a list of email
# addresses and/or filenames, separated by commas or newlines.  The files
# you reference should similarly be lists of email addresses and/or filenames.
SendTo: 

# Who should it appear this email is from?  This is normally of the form:
#   Some Body <somebody\@somewhere.com>
From: 

# Who should it appear this email is to?  You can leave this blank, or you
# can type "honest" to use the SendTo information.
To: 

# What should be the subject line of the email?
Subject: 

# Everything after "Body:" will be the body of the email message.
Body:

_eot_

    } else { 
	$template .= <<"_eot_";
# If you don't know how to use this, quit this editor and run:
#   kamaileon -tutorial

SendTo: 
From: 
To: 
Subject: 

Body:

_eot_

    }
    return $template;
}

